//#ifdef QT_QML_DEBUG
#include <QtQuick>
//#endif
#include <QMetaType>

#include <sailfishapp.h>

#include "dbusactivation.h"
#include "login.h"
#include "mxcdownloader.h"
#include "mxcimageprovider.h"
#include "settings.h"
#include "sync.h"

#include "models/roomlistmodel.h"

Q_DECLARE_METATYPE(std::vector<std::string>);

QDataStream &operator<<(QDataStream &out, const std::string &s) {
    out << QString::fromStdString(s);
    return out;
}
QDataStream &operator>>(QDataStream &in, std::string &s) {
    QString qs;
    in >> qs;
    s = qs.toStdString();
    return in;
}

namespace spoon {
Login *login;
RoomListModel *rooms;
RoomListSortProxy *roomsSorted;
Sync *sync;
MXCImageProvider *imageProvider;
MXCDownloader *downloader;

QQuickView *showApp(QGuiApplication *app) {
    spoon::imageProvider = new MXCImageProvider();
    auto v = SailfishApp::createView();

    v->engine()->addImageProvider("MXCImage", imageProvider);
    v->rootContext()->setContextProperty("login", login);
    v->rootContext()->setContextProperty("mxcDownloader", downloader);
    v->rootContext()->setContextProperty("rooms", roomsSorted);
    v->setSource(SailfishApp::pathTo("qml/harbour-spoon.qml"));
    v->show();

    return v;
}
} // namespace spoon

int main(int argc, char *argv[]) {
    if (DBusActivator::isRegistered()) {
        if (DBusActivator::raiseApp()) {
            return 0;
        } else {
            return 1;
        }
    }

    qmlRegisterUncreatableType<EventType>("dev.neko.spoon", 1, 0, "EventType", "Can't instantiate EventType");
    qRegisterMetaType<std::vector<mtx::events::collections::TimelineEvents>>(
        "std::vector<mtx::events::collections::TimelineEvents>");
    qRegisterMetaType<std::vector<mtx::events::collections::StateEvents>>(
        "std::vector<mtx::events::collections::StateEvents>");
    qRegisterMetaType<std::vector<std::string>>("std::vector<std::string>");
    qRegisterMetaType<QSharedPointer<Room>>("QSharedPointer<Room>");
    qRegisterMetaType<std::string>("std::string");
    qRegisterMetaTypeStreamOperators<std::string>("std::string");

    // SailfishApp::main() will display "qml/harbour-spoon.qml", if you need more
    // control over initialization, you can use:
    //
    //   - SailfishApp::application(int, char *[]) to get the QGuiApplication *
    //   - SailfishApp::createView() to get a new QQuickView * instance
    //   - SailfishApp::pathTo(QString) to get a QUrl to a resource file
    //   - SailfishApp::pathToMainQml() to get a QUrl to the main QML file
    //

    // return SailfishApp::main(argc, argv);

    // To display the view, call "show()" (will show fullscreen on device).
    // Set up QML engine.
    // qRegisterMetaType<Room>("Room");
    QScopedPointer<QGuiApplication> app(SailfishApp::application(argc, argv));
    app->setQuitOnLastWindowClosed(false);

    QScopedPointer<DBusActivator> dbusWatcher(new DBusActivator(app.data()));

    spoon::login = new Login();
    spoon::downloader = new MXCDownloader();
    spoon::imageProvider = new MXCImageProvider();
    spoon::rooms = new RoomListModel();
    spoon::sync = new Sync(*spoon::rooms, spoon::login);
    spoon::roomsSorted = new RoomListSortProxy();
    spoon::roomsSorted->setSourceModel(spoon::rooms);
    spoon::roomsSorted->setDynamicSortFilter(true);
    spoon::roomsSorted->sort(0);

    QObject::connect(spoon::login, &Login::loginSuccess, spoon::sync, &Sync::startSync, Qt::QueuedConnection);
    QObject::connect(spoon::login, &Login::loggedOut, spoon::sync, &Sync::logout, Qt::BlockingQueuedConnection);
    QObject::connect(spoon::login, &Login::loggedOut, spoon::rooms, &RoomListModel::clear,
                     Qt::BlockingQueuedConnection);
    QObject::connect(spoon::sync, &Sync::newRoom, spoon::rooms, &RoomListModel::addRoom, Qt::QueuedConnection);
    QObject::connect(spoon::sync, &Sync::roomUpdated, spoon::rooms, &RoomListModel::updateRoom, Qt::QueuedConnection);
    QObject::connect(app.data(), &QGuiApplication::aboutToQuit, spoon::sync, &Sync::exitSync, Qt::QueuedConnection);

    if (spoon::login->isLoggedIn())
        spoon::sync->startSync();

    // If you wish to publish your app on the Jolla harbour, it is recommended
    // that you prefix your internal namespaces with "harbour.".
    //
    // For details see:
    // https://harbour.jolla.com/faq#1.5.0
    // qmlRegisterType<DemoModel>("com.example", 1, 0, "DemoModel");

    // Start the application.
    DBusActivator::raiseApp();

    return app->exec();
}
